# Copyright (c) 2014-2018, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""iocage destroy module."""
import json
import os
import subprocess as su

import iocage_lib.ioc_json
import iocage_lib.ioc_stop
import libzfs


class IOCDestroy(object):

    """
    Destroy a jail's datasets and then if they have a RELEASE snapshot,
    destroy that as well.
    """

    def __init__(self):
        self.pool = iocage_lib.ioc_json.IOCJson().json_get_value("pool")
        self.iocroot = iocage_lib.ioc_json.IOCJson(
            self.pool).json_get_value("iocroot")
        self.zfs = libzfs.ZFS(history=True, history_prefix="<iocage>")
        self.ds = self.zfs.get_dataset

    @staticmethod
    def __stop_jails__(datasets, path=None, root=False, clean=False):
        if clean:
            jids, _ = su.Popen(["jls", "-n", "name", "jid",
                                "--libxo=json"], stdout=su.PIPE).communicate()
            jids = json.loads(jids)["jail-information"]["jail"]

            for j in jids:
                name = j["name"]
                jid = j["jid"]

                if "ioc-" in name:
                    su.Popen(["jail", "-r", jid]).communicate()

        for dataset in datasets.dependents:
            if "jails" not in dataset.name:
                continue

            if dataset.type != libzfs.DatasetType.FILESYSTEM:
                continue

            if dataset.properties["mountpoint"].value == 'legacy':
                continue

            # This is just to setup a replacement.
            path = path.replace("templates", "jails")

            try:
                uuid = dataset.name.partition(
                    f"{path}")[1].rsplit("/", 1)[1]
            except IndexError:
                # A RELEASE dataset
                return

            # We want the real path now.
            _path = dataset.properties["mountpoint"].value.replace("/root", "")
            # It gives us a string that says "none", not terribly
            # useful, fixing that.
            _path = _path if _path != "none" else None

            if (dataset.name.endswith(uuid) or root) and _path is not None:
                iocage_lib.ioc_stop.IOCStop(uuid, _path, silent=True)

    def __destroy_leftovers__(self, dataset, clean=False):
        """Removes parent datasets and logs."""
        uuid = dataset.name.rsplit("/root")[0].split("/")[-1]
        snapshot = False

        try:
            path = dataset.properties["mountpoint"].value

            if path.endswith("/root"):
                umount_path = path.rsplit("/root", 1)[0]
            else:
                umount_path = path

        except libzfs.ZFSException as err:
            # This is either not mounted or doesn't exist anymore,
            # we don't care either way.

            if err.code != libzfs.Error.NOENT:
                raise
            path = None
        except KeyError:
            # This is a snapshot
            path = None
            snapshot = True

        if path:
            try:
                os.remove(f"{self.iocroot}/log/{uuid}-console.log")
            except FileNotFoundError:
                pass

            # Dangling mounts are bad...mmkay?
            su.Popen(
                ["umount", "-afF", f"{umount_path}/fstab"],
                stderr=su.PIPE).communicate()
            su.Popen(
                ["umount", "-f", f"{umount_path}/root/dev/fd"],
                stderr=su.PIPE).communicate()
            su.Popen(
                ["umount", "-f", f"{umount_path}/root/dev"],
                stderr=su.PIPE).communicate()
            su.Popen(
                ["umount", "-f", f"{umount_path}/root/proc"],
                stderr=su.PIPE).communicate()
            su.Popen(
                ["umount", "-f", f"{umount_path}/root/compat/linux/proc"],
                stderr=su.PIPE).communicate()

        if not snapshot and \
                any(_type in dataset.name for _type
                    in ("jails", "templates", "releases")):
            # The jails parent won't show in the list.
            j_parent = self.ds(f"{dataset.name.replace('/root','')}")
            origin = j_parent.properties["origin"].value.rsplit("@", 1)[0]

            if origin != "":
                j_origin_snaps = self.ds(origin).snapshots_recursive
            else:
                j_origin_snaps = []

            j_dependents = j_parent.dependents

            for j_dependent in j_dependents:

                if j_dependent.type == libzfs.DatasetType.FILESYSTEM:
                    j_dependent.umount(force=True)

                j_dependent.delete()

            j_parent.umount(force=True)
            j_parent.delete()

            for snap in j_origin_snaps:
                p, n = snap.name.split("@")

                if n == uuid:
                    snap.delete()

    def __destroy_dataset__(self, dataset):
        """Destroys the given datasets and snapshots."""

        if dataset.type == libzfs.DatasetType.FILESYSTEM:
            origin = dataset.properties["origin"].value

            try:
                snap_dataset, snap = origin.split("@")
                self.ds(snap_dataset).destroy_snapshot(snap)
            except ValueError:
                pass  # This means we don't have an origin.

            dataset.umount(force=True)

        dataset.delete()

    def __destroy_parse_datasets__(self, path, clean=False, stop=True):
        """
        Parses the datasets before calling __destroy_dataset__ with each
        entry.
        """
        try:
            datasets = self.ds(path)
        except libzfs.ZFSException:
            # Dataset can't be found, we don't care

            return

        single = True if len(list(datasets.dependents)) == 0 else False
        dependents = datasets.dependents

        if single:
            # Is actually a single dataset.
            self.__destroy_dataset__(datasets)
        else:
            if "templates" in path or "release" in path:
                # This will tell __stop_jails__ to actually try stopping on
                # a /root
                root = True
            else:
                # Otherwise we only stop when the uuid is the last entry in
                # the jails path.
                root = False

            if stop:
                try:
                    self.__stop_jails__(datasets, path, root, clean=clean)
                except (RuntimeError, FileNotFoundError, SystemExit):
                    # If a bad or missing configuration for a jail, this will
                    # get in the way.
                    pass

            for dataset in dependents:
                try:
                    self.__destroy_dataset__(dataset)
                    self.__destroy_leftovers__(dataset, clean=clean)
                except libzfs.ZFSException as err:
                    # This is either not mounted or doesn't exist anymore,
                    # we don't care either way.

                    if err.code == libzfs.Error.NOENT:
                        continue
                    else:
                        raise

    def destroy_jail(self, path, clean=False):
        """
        A convenience wrapper to call __stop_jails__ and
         __destroy_parse_datasets__
        """
        dataset_type, uuid = path.rsplit("/")[-2:]

        if clean:
            self.__destroy_parse_datasets__(path, clean=clean)

            return

        try:
            iocage_lib.ioc_stop.IOCStop(uuid, path, silent=True)
        except (FileNotFoundError, RuntimeError, libzfs.ZFSException,
                SystemExit):
            # Broad exception as we don't care why this failed. iocage
            # may have been killed before configuration could be made,
            # it's meant to be nuked.
            pass

        try:
            self.__destroy_parse_datasets__(
                f"{self.pool}/iocage/{dataset_type}/{uuid}")
        except (libzfs.ZFSException, SystemExit):
            # The dataset doesn't exist, we don't care :)
            pass
